@@ -1,4 +1,4 @@
-from .base import MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, QueryError, ConnectError
+from data_diff.sqeleton.databases import MD5_HEXDIGITS, CHECKSUM_HEXDIGITS, QueryError, ConnectError
 
 from .postgresql import PostgreSQL
 from .mysql import MySQL
@@ -9,5 +9,8 @@ from .redshift import Redshift
 from .presto import Presto
 from .databricks import Databricks
 from .trino import Trino
+from .clickhouse import Clickhouse
+from .vertica import Vertica
+from .duckdb import DuckDB
 
-from .connect import connect_to_uri
+from ._connect import connect
